<?php

namespace Drupal\openapi_decoupled_toolbox\Form;

use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Class OpenApiDecoupledToolboxForm.
 */
class OpenApiDecoupledToolboxForm extends EntityForm {

  protected $entity_type_manager;

  /**
   * Class constructor.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager) {
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    // Instantiates this form class.
    return new static(
    // Load the service required to construct this class.
      $container->get('entity_type.manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    $form = parent::form($form, $form_state);
    $definitions = $this->entityTypeManager->getDefinitions();
    $options = [];
    foreach ($definitions as $entity_type => $def) {
      $view_modes = $this->entityTypeManager->getStorage('entity_view_display')
        ->loadByProperties(['targetEntityType' => $entity_type]);
      if (empty($view_modes)) {
        continue;
      }
      $options[$entity_type] = $def->getLabel()->__toString();
    }
    asort($options);
    /** @var  $open_api_decoupled_toolbox \Drupal\openapi_decoupled_toolbox\Entity\OpenApiDecoupledToolbox */
    $open_api_decoupled_toolbox = $this->entity;
    $form['entity_type'] = [
      '#type' => 'select',
      '#title' => t('Entity type'),
      '#required' => TRUE,
      '#default_value' => $open_api_decoupled_toolbox->getTargetEntityType() ?? '',
      '#size' => 20,
      '#options' => $options,
      '#ajax' => [
        'callback' => [$this, 'ajaxEntityType'],
        'wrapper' => 'ajax-entity-type',
        'method' => 'replaceWith',
      ],
    ];
    $form['container_entity_type'] = [
      '#type' => 'container',
      '#prefix' => '<div id="ajax-entity-type">',
      '#suffix' => '</div>',
    ];
    $form['container_entity_type']['bundle'] = [
      '#type' => 'hidden',
      '#default_value' => $open_api_decoupled_toolbox->getTargetBundle() ?? '',
      '#title' => t('Bundle'),
    ];
    $form['container_entity_type']['display'] = [
      '#type' => 'hidden',
      '#default_value' => $open_api_decoupled_toolbox->getTargetDisplay() ?? '',
      '#title' => t('Display'),
    ];
    /* You will need additional form elements for your custom properties. */

    $this->generateBundleForm($form, $form_state);
    $this->generateDisplayForm($form, $form_state);
    return $form;
  }

  /**
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @return bool
   */
  private function generateBundleForm(array &$form, FormStateInterface $form_state) {

    $definitions = $this->entityTypeManager->getDefinitions();
    $triggering_element = $form_state->getTriggeringElement();
    if ((empty($triggering_element['#name']) || $triggering_element['#name'] !== 'entity_type') && empty($form['container_entity_type']['bundle']['#default_value'])) {
      return FALSE;
    }
    $entity_type = $form_state->getValue('entity_type') ?? $form['entity_type']['#default_value'];
    if (empty($definitions[$entity_type])) {
      return FALSE;
    }
    $bundle_info = \Drupal::service('entity_type.bundle.info')
      ->getBundleInfo($entity_type);
    $options_bundle = [];
    foreach ($bundle_info as $bundle_name => $bundle) {
      $options_bundle[$bundle_name] = $bundle['label'];
    }
    $form['container_entity_type']['bundle']['#type'] = 'checkboxes';
    $form['container_entity_type']['bundle']['#required'] = TRUE;
    $form['container_entity_type']['bundle']['#multiple'] = TRUE;
    $form['container_entity_type']['bundle']['#size'] = 10;
    $form['container_entity_type']['bundle']['#options'] = $options_bundle;
    return TRUE;
  }

  private function generateDisplayForm(array &$form, FormStateInterface $form_state) {
    $triggering_element = $form_state->getTriggeringElement();
    if ((empty($triggering_element['#name']) || $triggering_element['#name'] !== 'entity_type') && empty($form['container_entity_type']['display']['#default_value'])) {
      return FALSE;
    }
    $entity_type = $form_state->getValue('entity_type') ?? $form['entity_type']['#default_value'];
    /** @var  $view_modes \Drupal\field_layout\Entity\FieldLayoutEntityViewDisplay[] */
    $view_modes = $this->entityTypeManager
      ->getStorage('entity_view_display')
      ->loadByProperties([
        'targetEntityType' => $entity_type,
      ]);
    if (empty($view_modes)) {
      return FALSE;
    }
    $options_display = [];
    foreach ($view_modes as  $display) {
      $options_display[$display->getMode()] = $display->label() ?? $display->getMode();
    }

    $form['container_entity_type']['display']['#type'] = 'checkboxes';
    $form['container_entity_type']['display']['#required'] = TRUE;
    $form['container_entity_type']['display']['#multiple'] = TRUE;
    $form['container_entity_type']['display']['#size'] = 10;
    $form['container_entity_type']['display']['#options'] = $options_display;
    return TRUE;
  }

  /**
   * Ajax callback of node bundle.
   *
   * {@inheritDoc}
   *
   * @SuppressWarnings(PHPMD.CamelCaseParameterName)
   * @SuppressWarnings(PHPMD.CamelCaseVariableName)
   */
  public static function ajaxEntityType(array &$form, FormStateInterface $form_state) {
    return $form['container_entity_type'];
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    $open_api_decoupled_toolbox = $this->entity;
    if ($open_api_decoupled_toolbox->isNew()) {
      $open_api_decoupled_toolbox->set('id', $open_api_decoupled_toolbox->getTargetEntityType() . '_' . implode('_', $open_api_decoupled_toolbox->getTargetBundle()));
      $open_api_decoupled_toolbox->set('label', $open_api_decoupled_toolbox->getTargetEntityType() . ' ' . implode(' ', $open_api_decoupled_toolbox->getTargetBundle()));
    }
    $status = $open_api_decoupled_toolbox->save();

    switch ($status) {
      case SAVED_NEW:
        $this->messenger()
          ->addMessage($this->t('Created the %label Open api decoupled toolbox.', [
            '%label' => $open_api_decoupled_toolbox->label(),
          ]));
        break;

      default:
        $this->messenger()
          ->addMessage($this->t('Saved the %label Open api decoupled toolbox.', [
            '%label' => $open_api_decoupled_toolbox->label(),
          ]));
    }
    $form_state->setRedirectUrl($open_api_decoupled_toolbox->toUrl('collection'));
  }

}
